@@ -81,6 +81,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
             use std::env;
             use tokio::task;
             use checkout_core::cloud_pubsub;
+            use checkout_core::schemars::JsonSchema;
             use checkout_core::dropshot;
             use checkout_core::dropshot::endpoint;
             use checkout_core::dropshot::RequestContext;
@@ -92,7 +93,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
             use checkout_core::checkout::Checkout;
             use checkout_core::payment::Payment;
             use checkout_core::payment::PaymentProcessor;
-            use checkout_core::payment::handlers::{process_handler, step_handler};
+            use checkout_core::payment::handlers::{process_handler, step_handler, callback_handler, cancel_handler};
             use checkout_core::customer::Contact;
             use checkout_core::customer::Address;
             use checkout_core::shipping::FulfillmentTypeSelection;
@@ -470,13 +471,14 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
             }
 
             // payment related endpoints
-            // struct CorrelationIdParam {
-            //     id: <super::#context_type_name<'_> as PaymentProcessor>::CorrelationID
-            // }
+            #[derive(JsonSchema)]
+            struct CorrelationIdParam {
+                correlation_id: <super::#context_type_name<'static> as PaymentProcessor>::CorrelationID
+            }
 
             #[endpoint {
                 method = PATCH,
-                path = "/payment/{id}/step_payment",
+                path = "/payment/{id}/step",
             }]
             pub async fn step_payment_endpoint(
                 req_ctx: Arc<RequestContext<ServerContext>>,
@@ -519,6 +521,91 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                 }
             }
+
+            #[endpoint {
+                method = PATCH,
+                path = "/payment/{correlation_id}/callback",
+            }]
+            pub async fn callback_payment_endpoint(
+                req_ctx: Arc<RequestContext<ServerContext>>,
+                path: Path<CorrelationIdParam>,
+                body: TypedBody<<super::#context_type_name<'_> as PaymentProcessor>::ProcessArgs>,
+            ) -> Result<HttpResponseOk<Payment<
+                <super::#context_type_name<'static> as PaymentProcessor>::CorrelationID,
+                <super::#context_type_name<'static> as PaymentProcessor>::PrivateData,
+                <super::#context_type_name<'static> as PaymentProcessor>::PublicData,
+                <super::#context_type_name<'static> as PaymentProcessor>::ProcessArgs
+            >>, HttpError> {
+                let path_params = path.into_inner();
+                let args = body.into_inner();
+                let internal_ctx = req_ctx
+                    .context()
+                    .internal_context
+                    .clone();
+
+                let mut ctx = req_ctx
+                    .context()
+                    .context_provider
+                    .new_context()
+                    .await
+                    .unwrap();
+
+                ctx.start_transaction().await.unwrap();
+                match callback_handler(&internal_ctx, &mut ctx, &path_params.correlation_id, args).await {
+                    Ok(co) => {
+                        ctx.commit_transaction().await.unwrap();
+                        Ok(HttpResponseOk(co))
+                    }
+                    Err(err) => {
+                        ctx.abort_transaction().await.unwrap();
+                        Err(HttpError {
+                            status_code: StatusCode::BAD_REQUEST,
+                            error_code: Some(err.code.clone()),
+                            external_message: err.message.clone(),
+                            internal_message: err.message.clone(),
+                        })
+                    }
+                }
+            }
+
+            #[endpoint {
+                method = PATCH,
+                path = "/payment/{id}/cancel",
+            }]
+            pub async fn cancel_payment_endpoint(
+                req_ctx: Arc<RequestContext<ServerContext>>,
+                path: Path<IdParam>,
+            ) -> Result<HttpResponseOk<Payment<
+                <super::#context_type_name<'static> as PaymentProcessor>::CorrelationID,
+                <super::#context_type_name<'static> as PaymentProcessor>::PrivateData,
+                <super::#context_type_name<'static> as PaymentProcessor>::PublicData,
+                <super::#context_type_name<'static> as PaymentProcessor>::ProcessArgs
+            >>, HttpError> {
+                let path_params = path.into_inner();
+                let mut ctx = req_ctx
+                    .context()
+                    .context_provider
+                    .new_context()
+                    .await
+                    .unwrap();
+
+                ctx.start_transaction().await.unwrap();
+                match cancel_handler(&mut ctx, &path_params.id).await {
+                    Ok(co) => {
+                        ctx.commit_transaction().await.unwrap();
+                        Ok(HttpResponseOk(co))
+                    }
+                    Err(err) => {
+                        ctx.abort_transaction().await.unwrap();
+                        Err(HttpError {
+                            status_code: StatusCode::BAD_REQUEST,
+                            error_code: Some(err.code.clone()),
+                            external_message: err.message.clone(),
+                            internal_message: err.message.clone(),
+                        })
+                    }
+                }
+            }
         }
 
         #[async_trait]
@@ -536,6 +623,8 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 api.register(checkout_server::update_fulfillment_type_endpoint).unwrap();
                 api.register(checkout_server::initiate_payment_endpoint).unwrap();
                 api.register(checkout_server::step_payment_endpoint).unwrap();
+                api.register(checkout_server::callback_payment_endpoint).unwrap();
+                api.register(checkout_server::cancel_payment_endpoint).unwrap();
 
                 api
             }
